#!/usr/bin/env python3
# -*- coding: utf-8 -*-
#
# === This file is part of Calamares - <https://calamares.io> ===
#
#   SPDX-FileCopyrightText: 2014 Rohan Garg <rohan@kde.org>
#   SPDX-FileCopyrightText: 2015 2019-2020, Philip Müller <philm@manjaro.org>
#   SPDX-FileCopyrightText: 2017 Alf Gaida <agaida@sidution.org>
#   SPDX-FileCopyrightText: 2019 Adriaan de Groot <groot@kde.org>
#   SPDX-License-Identifier: GPL-3.0-or-later
#
#   Calamares is Free Software: see the License-Identifier above.
#
import libcalamares
from libcalamares.utils import debug, target_env_call
import os
from collections import OrderedDict

import gettext
_ = gettext.translation("calamares-python",
                        localedir=libcalamares.utils.gettext_path(),
                        languages=libcalamares.utils.gettext_languages(),
                        fallback=True).gettext


def pretty_name():
    return _("Configuring mkinitcpio.")


def detect_plymouth():
    """
    Checks existence (runnability) of plymouth in the target system.

    @return True if plymouth exists in the target, False otherwise
    """
    # Used to only check existence of path /usr/bin/plymouth in target
    return target_env_call(["sh", "-c", "which plymouth"]) == 0


class cpuinfo(object):
    """
    Object describing the current CPU's characteristics. It may be
    be considered a named tuple, there's no behavior here.

    Fields in the object:
        - is_intel (if it's definitely an Intel CPU)
        - is_amd (if it's definitely an AMD CPU)
        - number_of_cores
    It is possible for both is_* fields to be False.
    """
    def __init__(self):
        self.is_intel = False
        self.is_amd = False
        self.number_of_cores = 0

        cpu = self._cpuinfo()
        if 'vendor_id' in cpu['proc0']:
            self.is_intel = cpu['proc0']['vendor_id'].lower() == "genuineintel"
            self.is_amd = cpu['proc0']['vendor_id'].lower() == "authenticamd"
        self.number_of_cores = len(cpu)

    @staticmethod
    def _cpuinfo():
        """
        Return the information in /proc/cpuinfo as a dictionary in the following
        format:

        cpu_info['proc0']={...}
        cpu_info['proc1']={...}
        """
        cpu_info = OrderedDict()
        procinfo = OrderedDict()

        nprocs = 0

        with open('/proc/cpuinfo') as cpuinfo_file:
            for line in cpuinfo_file:
                if not line.strip():
                    # end of one processor
                    cpu_info["proc{!s}".format(nprocs)] = procinfo
                    nprocs += 1
                    # Reset
                    procinfo = OrderedDict()
                else:
                    if len(line.split(':')) == 2:
                        splitted_line = line.split(':')[1].strip()
                        procinfo[line.split(':')[0].strip()] = splitted_line
                    else:
                        procinfo[line.split(':')[0].strip()] = ''

        return cpu_info


def get_host_initcpio():
    """
    Reads the host system mkinitcpio.conf and returns all
    the lines from that file, or an empty list if it does
    not exist.
    """
    hostfile = libcalamares.job.configuration.get("source", None) or "/etc/mkinitcpio.conf"
    try:
        with open(hostfile, "r") as mkinitcpio_file:
            mklins = [x.strip() for x in mkinitcpio_file.readlines()]
    except FileNotFoundError:
        libcalamares.utils.debug(f"Could not open host file {hostfile}")
        mklins = []

    return mklins


def write_mkinitcpio_lines(hooks, modules, files, binaries, root_mount_point):
    """
    Set up mkinitcpio.conf.

    :param hooks:
    :param modules:
    :param files:
    :param root_mount_point:
    """
    mklins = get_host_initcpio()

    target_path = os.path.join(root_mount_point, "etc/mkinitcpio.conf")
    with open(target_path, "w") as mkinitcpio_file:
        for line in mklins:
            # Replace HOOKS, MODULES, BINARIES and FILES lines with what we
            # have found via find_initcpio_features()
            if line.startswith("HOOKS"):
                line = f"HOOKS=({str.join(' ', hooks)})"
            elif line.startswith("BINARIES"):
                line = f"BINARIES=({str.join(' ', binaries)})"
            elif line.startswith("MODULES"):
                line = f"MODULES=({str.join(' ', modules)})"
            elif line.startswith("FILES"):
                line = f"FILES=({str.join(' ', files)})"
            mkinitcpio_file.write(line + "\n")


def find_initcpio_features(partitions, root_mount_point):
    """
    Returns a tuple (hooks, modules, files) needed to support
    the given @p partitions (filesystems types, encryption, etc)
    in the target.

    :param partitions: (from GS)
    :param root_mount_point: (from GS)

    :return 3-tuple of lists
    """
    hooks = [
        "autodetect",
        "microcode",
        "kms",
        "modconf",
        "block",
        "keyboard",
    ]

    systemd_hook_allowed = libcalamares.job.configuration.get("useSystemdHook", False)

    use_systemd = systemd_hook_allowed and target_env_call(["sh", "-c", "which systemd-cat"]) == 0

    if use_systemd:
        hooks.insert(0, "systemd")
        hooks.append("sd-vconsole")
    else:
        hooks.insert(0, "udev")
        hooks.insert(0, "base")
        hooks.append("keymap")
        hooks.append("consolefont")

    hooks_map = libcalamares.job.configuration.get("hooks", None)
    if not hooks_map:
        hooks_map = dict()
    hooks_prepend = hooks_map.get("prepend", None) or []
    hooks_append = hooks_map.get("append", None) or []
    hooks_remove = hooks_map.get("remove", None) or []

    modules = []
    files = []
    binaries = []

    swap_uuid = ""
    uses_btrfs = False
    uses_zfs = False
    uses_lvm2 = False
    encrypt_hook = False
    openswap_hook = False
    unencrypted_separate_boot = False

    # It is important that the plymouth hook comes before any encrypt hook
    if detect_plymouth():
        hooks.append("plymouth")

    for partition in partitions:
        if partition["fs"] == "linuxswap" and not partition.get("claimed", None):
            # Skip foreign swap
            continue

        if partition["fs"] == "linuxswap":
            swap_uuid = partition["uuid"]
            if "luksMapperName" in partition:
                openswap_hook = True

        if partition["fs"] == "btrfs":
            uses_btrfs = True

        # In addition to checking the filesystem, check to ensure that zfs is enabled
        if partition["fs"] == "zfs" and libcalamares.globalstorage.contains("zfsPoolInfo"):
            uses_zfs = True

        if "lvm2" in partition["fs"]:
            uses_lvm2 = True

        if partition["mountPoint"] == "/" and "luksMapperName" in partition:
            encrypt_hook = True

        if partition["mountPoint"] == "/boot" and "luksMapperName" not in partition:
            unencrypted_separate_boot = True

        if partition["mountPoint"] == "/usr":
            hooks.append("usr")

    if encrypt_hook:
        if use_systemd:
            hooks.append("sd-encrypt")
        else:
            hooks.append("encrypt")
        crypto_file = "crypto_keyfile.bin"
        if not unencrypted_separate_boot and os.path.isfile(os.path.join(root_mount_point, crypto_file)):
            files.append(f"/{crypto_file}")

    if uses_lvm2:
        hooks.append("lvm2")

    if uses_zfs:
        hooks.append("zfs")

    if swap_uuid != "":
        if encrypt_hook and openswap_hook:
            hooks.extend(["openswap"])
        hooks.extend(["resume", "filesystems"])
    else:
        hooks.extend(["filesystems"])

    if uses_btrfs:
        modules.append("crc32c-intel" if cpuinfo().is_intel else "crc32c")
    else:
        hooks.append("fsck")

    # Modify according to the keys in the configuration
    hooks = [h for h in (hooks_prepend + hooks + hooks_append) if h not in hooks_remove]

    return hooks, modules, files, binaries


def run():
    """
    Calls routine with given parameters to modify "/etc/mkinitcpio.conf".

    :return:
    """
    partitions = libcalamares.globalstorage.value("partitions")
    root_mount_point = libcalamares.globalstorage.value("rootMountPoint")

    if not partitions:
        libcalamares.utils.warning(f"partitions are empty, {partitions}")
        return (_("Configuration Error"),
                _("No partitions are defined for <pre>initcpiocfg</pre>."))
    if not root_mount_point:
        libcalamares.utils.warning(f"rootMountPoint is empty, {root_mount_point}")
        return (_("Configuration Error"),
                _("No root mount point for <pre>initcpiocfg</pre>."))

    hooks, modules, files, binaries = find_initcpio_features(partitions, root_mount_point)
    write_mkinitcpio_lines(hooks, modules, files, binaries, root_mount_point)

    return None



#!/usr/bin/env python3

import os
import subprocess
import shutil

import libcalamares
from libcalamares.utils import gettext_path, gettext_languages

import gettext

_translation = gettext.translation("calamares-python",
                                   localedir=gettext_path(),
                                   languages=gettext_languages(),
                                   fallback=True)
_ = _translation.gettext
_n = _translation.ngettext

custom_status_message = None
name = "Prepare for bootloader"
user_output = False


def pretty_name():
    return _(name)


def pretty_status_message():
    if custom_status_message is not None:
        return custom_status_message


def is_resume_needed():
    partitions = libcalamares.globalstorage.value("partitions")
    for partition in partitions:
        if partition["fs"] == "linuxswap":
            return True

    return False


def get_local_packages(packages):
    try:
        package_location = libcalamares.job.configuration["packageLocation"]
    except KeyError:
        return "Configuration Error", "No package location defined in config"

    if not os.path.exists(package_location):
        return ("Package location missing",
                f"{package_location} is not currently available")

    try:
        installation_root_path = libcalamares.globalstorage.value("rootMountPoint")
    except KeyError:
        libcalamares.utils.warning('Global storage value "rootMountPoint" missing')

    package_files = []
    for root, dirs, files in os.walk(os.path.join(installation_root_path, package_location.lstrip('/'))):
        for file in files:
            for package in packages:
                if file.startswith(package + "-") and file.endswith(".zst"):
                    package_files.append(os.path.join(package_location, file))

    return package_files


def run_dracut(installation_root_path):
    kernel_search_path = "/usr/lib/modules"

    # find all the installed kernels and run dracut
    for root, dirs, files in os.walk(os.path.join(installation_root_path, kernel_search_path.lstrip('/'))):
        for file in files:
            if file == "pkgbase":
                kernel_version = os.path.basename(root)
                # run dracut
                pkgbase_location = os.path.join(root, file)
                with open(pkgbase_location, 'r') as pkgbase_file:
                    kernel_suffix = pkgbase_file.read().rstrip()
                try:
                    libcalamares.utils.target_env_process_output(["dracut", "--force", "--hostonly",
                                                                  "--no-hostonly-cmdline",
                                                                  f"/boot/initramfs-{kernel_suffix}.img",
                                                                  kernel_version])
                    libcalamares.utils.target_env_process_output(["dracut", "--force", "--no-hostonly",
                                                                  f"/boot/initramfs-{kernel_suffix}-fallback.img",
                                                                  kernel_version])
                except subprocess.CalledProcessError as cpe:
                    libcalamares.utils.warning(f"dracut failed with error: {cpe.stderr}")

                kernel_name = f"vmlinuz-{kernel_suffix}"
                # copy kernel to boot
                shutil.copy2(os.path.join(root, "vmlinuz"), os.path.join(installation_root_path, "boot", kernel_name))


def run():
    if not libcalamares.job.configuration:
        return "No configuration found", "Aborting due to missing configuration"

    try:
        gs_name = libcalamares.job.configuration["gsName"]
    except KeyError:
        return "Missing global storage value", "gsname not found in configuration file"

    try:
        offline = libcalamares.job.configuration["offline"]
    except KeyError:
        offline = False
        pass

    bootloaders = libcalamares.job.configuration.get("bootloader", [])

    if libcalamares.globalstorage.contains(gs_name):
        bootloader_name = libcalamares.globalstorage.value(gs_name)
    else:
        return f"Key missing", f"Failed to find {gs_name} in global storage"

    try:
        installation_root_path = libcalamares.globalstorage.value("rootMountPoint")
    except KeyError:
        libcalamares.utils.warning('Global storage value "rootMountPoint" missing')

    packages = None

    for bootloader in bootloaders:
        try:
            if bootloader["name"].casefold() == bootloader_name.casefold():
                packages = bootloader["packages"]
        except KeyError:
            return f"Configuration error", f"Missing key 'name' in configuration"

    # Add the resume module for dracut
    if is_resume_needed():
        dracut_file_path = os.path.join(installation_root_path , "etc/dracut.conf.d/resume.conf")
        resume_line = 'add_dracutmodules+=" resume "'
        with open(dracut_file_path, 'w') as dracut_resume:
            dracut_resume.write(resume_line + "\n")

    # install packages
    if offline:
        package_files = get_local_packages(packages)
        if package_files is not None:
            try:
                libcalamares.utils.target_env_process_output(["pacman", "--noconfirm", "-U"] + package_files)
            except subprocess.CalledProcessError as cpe:
                return f"Failed to install packages for {bootloader_name}", f"The install failed with error: {cpe.stderr}"
    else:
        if packages is not None:
            try:
                libcalamares.utils.target_env_process_output(["pacman", "--noconfirm", "-S"] + packages)
            except subprocess.CalledProcessError as cpe:
                return f"Failed to install packages for {bootloader_name}", f"The install failed with error: {cpe.stderr}"

    # Run dracut unless we are using systemd-boot since kernel-install handles that
    if bootloader_name.casefold().strip() != "systemd-boot":
        run_dracut(installation_root_path)

    return None



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
#   SPDX-License-Identifier: GPL-3.0-or-later
#

import fileinput
import os
import shutil
import subprocess
import stat
import tempfile

import libcalamares

from libcalamares.utils import check_target_env_call
from libcalamares.utils import check_target_env_output

import gettext

_ = gettext.translation("calamares-python",
                        localedir=libcalamares.utils.gettext_path(),
                        languages=libcalamares.utils.gettext_languages(),
                        fallback=True).gettext


def pretty_name():
    return _("Install bootloader.")


def copytree(source_dir, dest_dir):
    """
    Recursively copy source_dir to dest_dir.  This is basically a work-around for the fact that native Python copytree
    can't handle the situation where dest_dir already exists
    """
    if not os.path.exists(dest_dir):
        os.makedirs(dest_dir)
    for item in os.listdir(source_dir):
        s = os.path.join(source_dir, item)
        d = os.path.join(dest_dir, item)
        if os.path.isdir(s):
            copytree(s, d)
        else:
            shutil.copy2(s, d)


def handle_systemdboot(efi_directory):
    """
    For systemd-boot, we copy the Microsoft EFI entry to the EFI partition where systemd-boot is installed
    """

    # Get the root mount point and build the full path to the mounted ESP
    try:
        installation_root_path = libcalamares.globalstorage.value("rootMountPoint")
    except KeyError:
        libcalamares.utils.warning('Global storage value "rootMountPoint" missing')

    install_efi_directory = installation_root_path + efi_directory

    # espList holds the list of EFI partitions from the partition module
    try:
        efi_partition_list = libcalamares.globalstorage.value("espList")
    except KeyError:
        libcalamares.utils.warning("No ESP list in global storage")
        return None

    # Get the partitions from global storage
    try:
        partitions = libcalamares.globalstorage.value("partitions")
    except KeyError:
        libcalamares.utils.warning("partitions missing from global storage")
        return None

    # Iterate over the partitions and search for an EFI partition containing the Microsoft data
    for partition in partitions:
        try:
            if partition["device"] in efi_partition_list and partition["mountPoint"] != efi_directory:
                libcalamares.utils.debug(f"Found foreign efi at {partition['device']}")
                temp_path = tempfile.mkdtemp()
                libcalamares.utils.debug(f"Attempting to mount {partition['device'].strip()} at {temp_path}")
                libcalamares.utils.host_env_process_output(["mount", partition["device"].strip(), temp_path])
                source_path = os.path.join(temp_path, "EFI", "Microsoft")
                if os.path.isdir(source_path):
                    target_path = os.path.join(install_efi_directory, "EFI", "Microsoft")
                    copytree(source_path, target_path)
                libcalamares.utils.host_env_process_output(["umount", temp_path])
        except KeyError:
            pass
        except subprocess.CalledProcessError:
            libcalamares.utils.warning("Failed to mount foreign EFI dir")
            pass
        except shutil.Error:
            libcalamares.utils.warning("Failed to copy Windows files from foreign EFI dir")
            pass


def enable_osprober(installation_root_path, enable):
    """
    Enabled or disabled os-prober in the target based on the value of the enable parameter
    """

    # Update the config in the file to enable or disable os-prober
    for line in fileinput.input(os.path.join(installation_root_path, "etc/default/grub"), inplace=True):
        line = line.strip()
        if line.startswith("#GRUB_DISABLE_OS_PROBER=false") and enable:
            print(line.lstrip("#"))
        elif line.startswith("GRUB_DISABLE_OS_PROBER=false") and not enable:
            print("#" + line)
        else:
            print(line)


def write_grub_config(installation_root_path, filename, entry):
    """
    Creates a custom grub entry with the data found in "entry" with the name "filename"
    """

    conf_path = os.path.join(installation_root_path, "etc", "grub.d", filename)
    with open(conf_path, 'w') as conf_file:
        conf_file.write("#!/bin/sh\n")
        conf_file.write("exec tail -n +3 $0\n\n")
        conf_file.writelines(entry)

    try:
        os.chmod(conf_path, stat.S_IRWXU | stat.S_IRGRP | stat.S_IXGRP | stat.S_IROTH | stat.S_IXOTH)
    except os.error as e:
        libcalamares.utils.warning(f"Failed to chmod grub config, the error was: {e}")
        
    return None


def handle_grub():
    """
    Enables os-prober and checks for a windows entry.  If it finds one, it is written to /etc/grub.d/45_eos_windows.
    """

    # Get the root mount point
    try:
        installation_root_path = libcalamares.globalstorage.value("rootMountPoint")
    except KeyError:
        libcalamares.utils.warning('Global storage value "rootMountPoint" missing')
        return None

    enable_osprober(installation_root_path, True)

    mkconfig_output = check_target_env_output(["grub-mkconfig"])

    # Find the contents of the entry
    found_osprober = False
    found_windows = False
    for line in mkconfig_output.splitlines():
        if line.strip() == "### BEGIN /etc/grub.d/30_os-prober ###":
            found_osprober = True
        elif "menuentry" in line and "Windows" in line and found_osprober:
            found_windows = True
            entry = [f"{line}\n"]
        elif found_windows and found_osprober:
            entry.append(f"{line}\n")
            if line.strip().startswith("}"):
                break

    if found_windows:
        write_grub_config(installation_root_path, "45_eos_windows", entry)
    else:
        libcalamares.utils.debug("No Windows entry found by os-prober")

    # We are done, disable os-prober
    enable_osprober(installation_root_path, False)
    return None


def run():
    fw_type = libcalamares.globalstorage.value("firmwareType")

    partitions = libcalamares.globalstorage.value("partitions")
    efi_directory = libcalamares.globalstorage.value("efiSystemPartition")
    if fw_type == "efi":
        esp_found = [p for p in partitions if p["mountPoint"] == efi_directory]
        if not esp_found:
            libcalamares.utils.warning(f"EFI system, but nothing mounted on {efi_directory}")
            return None

    # Get the boot loader selection from global storage if it is set in the config file
    try:
        gs_name = libcalamares.job.configuration["bootLoaderVar"]
        boot_loader = libcalamares.globalstorage.value(gs_name)
    except KeyError:
        libcalamares.utils.warning(
            f"Specified global storage value not found in global storage")
        return None

    # If the user has selected not to install bootloader, bail out here
    if boot_loader.casefold() == "none":
        libcalamares.utils.debug("No bootloader selecting, skipping Windows boot entry creation")
        return None

    if boot_loader == "systemd-boot":
        handle_systemdboot(efi_directory)
    elif boot_loader == "grub":
        handle_grub()
    else:
        libcalamares.utils.debug(f"No manual Windows entry creation available for {boot_loader}")

    return None



#!/usr/bin/env python3

import libcalamares



def run():
    """
    Sets global storage variable for ARM installs

    """

    libcalamares.globalstorage.insert("arm_install", True)

    libcalamares.job.setprogress(1.0)

    return None



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
import os
import shutil

import libcalamares
from libcalamares.utils import *


def run():
    """ Copy kernel and install scripts to target system.

    :return:
    """

    kernel_path = "/run/archiso/bootmnt/arch/boot/x86_64/"
    kernel_root = "vmlinuz"
    ucode_tail = "ucode.img"

    root_mount_point = libcalamares.globalstorage.value("rootMountPoint")
    if not root_mount_point:
        return ("No mount point for root partition in globalstorage",
                "globalstorage does not contain a \"rootMountPoint\" key, "
                "doing nothing")

    if not os.path.exists(root_mount_point):
        return ("Bad mount point for root partition in globalstorage",
                "globalstorage[\"rootMountPoint\"] is \"{}\", which does not "
                "exist, doing nothing".format(root_mount_point))

    try:
        # Copy any kernels and ucode available to /boot in the target
        for file in os.listdir(kernel_path):
            if file.startswith(kernel_root) or file.endswith(ucode_tail):
                shutil.copy2(file, os.path.join(root_mount_point, "boot", os.path.basename(file)))
    except Exception as e:
        return "File copy failed", "kernel-copy failed to copy file with error " + format(e)

    return None



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
#
# === This file is part of Calamares - <https://calamares.io> ===
#
#   SPDX-FileCopyrightText: 2014 Teo Mrnjavac <teo@kde.org>
#   SPDX-FileCopyrightText: 2014 Daniel Hillenbrand <codeworkx@bbqlinux.org>
#   SPDX-FileCopyrightText: 2014 Philip Müller <philm@manjaro.org>
#   SPDX-FileCopyrightText: 2017 Alf Gaida <agaida@siduction.org>
#   SPDX-FileCopyrightText: 2019 Kevin Kofler <kevin.kofler@chello.at>
#   SPDX-FileCopyrightText: 2020 Adriaan de Groot <groot@kde.org>
#   SPDX-FileCopyrightText: 2020 Gabriel Craciunescu <crazy@frugalware.org>
#   SPDX-License-Identifier: GPL-3.0-or-later
#
#   Calamares is Free Software: see the License-Identifier above.
#

import os
import re
import shutil
import subprocess
import sys
import tempfile

import libcalamares

import gettext
_ = gettext.translation("calamares-python",
                        localedir=libcalamares.utils.gettext_path(),
                        languages=libcalamares.utils.gettext_languages(),
                        fallback=True).gettext

def pretty_name():
    return _("Filling up filesystems.")

# This is going to be changed from various methods
status = pretty_name()

def pretty_status_message():
    return status

class UnpackEntry:
    """
    Extraction routine using rsync.

    :param source:
    :param sourcefs:
    :param destination:
    """
    __slots__ = ('source', 'sourcefs', 'destination', 'copied', 'total', 'exclude', 'excludeFile',
                 'mountPoint', 'weight')

    def __init__(self, source, sourcefs, destination):
        """
        @p source is the source file name (might be an image file, or
            a directory, too)
        @p sourcefs is a type indication; "file" is special, as is
            "squashfs".
        @p destination is where the files from the source go. This is
            **already** prefixed by rootMountPoint, so should be a
            valid absolute path within the host system.

        The members copied and total are filled in by the copying process.
        """
        self.source = source
        self.sourcefs = sourcefs
        self.destination = destination
        self.exclude = None
        self.excludeFile = None
        self.copied = 0
        self.total = 0
        self.mountPoint = None
        self.weight = 1

    def is_file(self):
        return self.sourcefs == "file"

    def do_count(self):
        """
        Counts the number of files this entry has.
        """
        # Need a name we can use like a global
        class counter(object):
            count = 0
        def cb_count(s):
            counter.count += 1

        if self.sourcefs == "squashfs":
            libcalamares.utils.host_env_process_output(["unsquashfs", "-l", self.source], cb_count)

        elif self.sourcefs == "ext4":
            libcalamares.utils.host_env_process_output(["find", self.mountPoint, "-type", "f"], cb_count)

        elif self.is_file():
            # Hasn't been mounted, copy directly; find handles both
            # files and directories.
            libcalamares.utils.host_env_process_output(["find", self.source, "-type", "f"], cb_count)

        self.total = counter.count
        return self.total

    def do_mount(self, base):
        """
        Mount given @p entry as loop device underneath @p base

        A *file* entry (e.g. one with *sourcefs* set to *file*)
        is not mounted and just ignored.

        :param base: directory to place all the mounts in.

        :returns: None, but throws if the mount failed
        """
        imgbasename = os.path.splitext(
            os.path.basename(self.source))[0]
        imgmountdir = os.path.join(base, imgbasename)
        os.makedirs(imgmountdir, exist_ok=True)

        # This is where it *would* go (files bail out before actually mounting)
        self.mountPoint = imgmountdir

        if self.is_file():
            return

        if os.path.isdir(self.source):
            r = libcalamares.utils.mount(self.source, imgmountdir, "", "--bind")
        elif os.path.isfile(self.source):
            r = libcalamares.utils.mount(self.source, imgmountdir, self.sourcefs, "loop")
        else: # self.source is a device
            r = libcalamares.utils.mount(self.source, imgmountdir, self.sourcefs, "")

        if r != 0:
            libcalamares.utils.debug("Failed to mount '{}' (fs={}) (target={})".format(self.source, self.sourcefs, imgmountdir))
            raise subprocess.CalledProcessError(r, "mount")


ON_POSIX = 'posix' in sys.builtin_module_names


def global_excludes():
    """
    List excludes for rsync.
    """
    lst = []
    extra_mounts = libcalamares.globalstorage.value("extraMounts")
    if extra_mounts is None:
        extra_mounts = []

    for extra_mount in extra_mounts:
        mount_point = extra_mount["mountPoint"]

        if mount_point:
            lst.extend(['--exclude', mount_point + '/'])

    return lst

def file_copy(source, entry, progress_cb):
    """
    Extract given image using rsync.

    :param source: Source file. This may be the place the entry's
        image is mounted, or if it's a single file, the entry's source value.
    :param entry: The UnpackEntry being copied.
    :param progress_cb: A callback function for progress reporting.
        Takes a number and a total-number.
    """
    import time

    dest = entry.destination

    # `source` *must* end with '/' otherwise a directory named after the source
    # will be created in `dest`: ie if `source` is "/foo/bar" and `dest` is
    # "/dest", then files will be copied in "/dest/bar".
    if not source.endswith("/") and not os.path.isfile(source):
        source += "/"

    num_files_total_local = 0
    num_files_copied = 0  # Gets updated through rsync output

    args = ['rsync', '-aHAXSr', '--filter=-x trusted.overlay.*']
    args.extend(global_excludes())
    if entry.excludeFile:
        args.extend(["--exclude-from=" + entry.excludeFile])
    if entry.exclude:
        for f in entry.exclude:
            args.extend(["--exclude", f])
    args.extend(['--progress', source, dest])

    # last_num_files_copied trails num_files_copied, and whenever at least 107 more
    # files (file_count_chunk) have been copied, progress is reported and
    # last_num_files_copied is updated. The chunk size isn't "tidy"
    # so that all the digits of the progress-reported number change.
    #
    file_count_chunk = 107

    class counter(object):
        last_num_files_copied = 0
        last_timestamp_reported = time.time()
        last_total_reported = 0

    def output_cb(line):
        # rsync outputs progress in parentheses. Each line will have an
        # xfer and a chk item (either ir-chk or to-chk) as follows:
        #
        # - xfer#x => Interpret it as 'file copy try no. x'
        # - ir-chk=x/y, where:
        #     - x = number of files yet to be checked
        #     - y = currently calculated total number of files.
        # - to-chk=x/y, which is similar and happens once the ir-chk
        #   phase (collecting total files) is over.
        #
        # If you're copying directory with some links in it, the xfer#
        # might not be a reliable counter (for one increase of xfer, many
        # files may be created).
        m = re.findall(r'xfr#(\d+), ..-chk=(\d+)/(\d+)', line)

        if m:
            # we've got a percentage update
            num_files_remaining = int(m[0][1])
            num_files_total_local = int(m[0][2])
            # adjusting the offset so that progressbar can be continuesly drawn
            num_files_copied = num_files_total_local - num_files_remaining

            now = time.time()
            if (num_files_copied - counter.last_num_files_copied >= file_count_chunk) or (now - counter.last_timestamp_reported > 0.5):
                counter.last_num_files_copied = num_files_copied
                counter.last_timestamp_reported = now
                counter.last_total_reported = num_files_total_local
                progress_cb(num_files_copied, num_files_total_local)

    try:
        returncode = libcalamares.utils.host_env_process_output(args, output_cb)
    except subprocess.CalledProcessError as e:
        returncode = e.returncode

    progress_cb(counter.last_num_files_copied, counter.last_total_reported)  # Push towards 100%

    # Mark this entry as really done
    entry.copied = entry.total

    # 23 is the return code rsync returns if it cannot write extended
    # attributes (with -X) because the target file system does not support it,
    # e.g., the FAT EFI system partition. We need -X because distributions
    # using file system capabilities and/or SELinux require the extended
    # attributes. But distributions using SELinux may also have SELinux labels
    # set on files under /boot/efi, and rsync complains about those. The only
    # clean way would be to split the rsync into one with -X and
    # --exclude /boot/efi and a separate one without -X for /boot/efi, but only
    # if /boot/efi is actually an EFI system partition. For now, this hack will
    # have to do. See also:
    # https://bugzilla.redhat.com/show_bug.cgi?id=868755#c50
    # for the same issue in Anaconda, which uses a similar workaround.
    if returncode != 0 and returncode != 23:
        libcalamares.utils.warning("rsync failed with error code {}.".format(returncode))
        return _("rsync failed with error code {}.").format(returncode)

    return None


class UnpackOperation:
    """
    Extraction routine using unsquashfs.

    :param entries:
    """

    def __init__(self, entries):
        self.entries = entries
        self.entry_for_source = dict((x.source, x) for x in self.entries)
        self.total_weight = sum([e.weight for e in entries])

    def report_progress(self):
        """
        Pass progress to user interface
        """
        progress = float(0)

        current_total = 0
        current_done = 0  # Files count in the current entry
        complete_count = 0
        complete_weight = 0  # This much weight already finished
        for entry in self.entries:
            if entry.total == 0:
                # Total 0 hasn't counted yet
                continue
            if entry.total == entry.copied:
                complete_weight += entry.weight
                complete_count += 1
            else:
                # There is at most *one* entry in-progress
                current_total = entry.total
                current_done = entry.copied
                complete_weight += entry.weight * ( 1.0 * current_done ) / current_total
                break

        if current_total > 0:
            progress = ( 1.0 * complete_weight ) / self.total_weight

        global status
        status = _("Unpacking image {}/{}, file {}/{}").format((complete_count+1), len(self.entries), current_done, current_total)
        libcalamares.job.setprogress(progress)

    def run(self):
        """
        Extract given image using unsquashfs.

        :return:
        """
        global status
        source_mount_path = tempfile.mkdtemp()

        try:
            complete = 0
            for entry in self.entries:
                status = _("Starting to unpack {}").format(entry.source)
                libcalamares.job.setprogress( ( 1.0 * complete ) / len(self.entries) )
                entry.do_mount(source_mount_path)
                entry.do_count()  # Fill in the entry.total

                self.report_progress()
                error_msg = self.unpack_image(entry, entry.mountPoint)

                if error_msg:
                    return (_("Failed to unpack image \"{}\"").format(entry.source),
                            error_msg)
                complete += 1

            return None
        finally:
            shutil.rmtree(source_mount_path, ignore_errors=True, onerror=None)


    def unpack_image(self, entry, imgmountdir):
        """
        Unpacks image.

        :param entry:
        :param imgmountdir:
        :return:
        """
        def progress_cb(copied, total):
            """ Copies file to given destination target.

            :param copied:
            """
            entry.copied = copied
            if total > entry.total:
                entry.total = total
            self.report_progress()

        try:
            if entry.is_file():
                source = entry.source
            else:
                source = imgmountdir

            return file_copy(source, entry, progress_cb)
        finally:
            if not entry.is_file():
                subprocess.check_call(["umount", "-l", imgmountdir])


def get_supported_filesystems_kernel():
    """
    Reads /proc/filesystems (the list of supported filesystems
    for the current kernel) and returns a list of (names of)
    those filesystems.
    """
    PATH_PROCFS = '/proc/filesystems'

    if os.path.isfile(PATH_PROCFS) and os.access(PATH_PROCFS, os.R_OK):
        with open(PATH_PROCFS, 'r') as procfile:
            filesystems = procfile.read()
            filesystems = filesystems.replace(
                "nodev", "").replace("\t", "").splitlines()
            return filesystems

    return []


def get_supported_filesystems():
    """
    Returns a list of all the supported filesystems
    (valid values for the *sourcefs* key in an item.
    """
    return ["file"] + get_supported_filesystems_kernel()


def repair_root_permissions(root_mount_point):
    """
    If the / of the system gets permission 777, change it down
    to 755. Any other permission is left alone. This
    works around standard behavior from squashfs where
    permissions are (easily, accidentally) set to 777.
    """
    existing_root_mode = os.stat(root_mount_point).st_mode & 0o777
    if existing_root_mode == 0o777:
        try:
            os.chmod(root_mount_point, 0o755)  # Want / to be rwxr-xr-x
        except OSError as e:
            libcalamares.utils.warning("Could not set / to safe permissions: {}".format(e))
            # But ignore it


def extract_weight(entry):
    """
    Given @p entry, a dict representing a single entry in
    the *unpack* list, returns its weight (1, or whatever is
    set if it is sensible).
    """
    w =  entry.get("weight", None)
    if w:
        try:
            wi = int(w)
            return wi if wi > 0 else 1
        except ValueError:
            libcalamares.utils.warning("*weight* setting {!r} is not valid.".format(w))
        except TypeError:
            libcalamares.utils.warning("*weight* setting {!r} must be number.".format(w))
    return 1


def run():
    """
    Unsquash filesystem.
    """
    root_mount_point = libcalamares.globalstorage.value("rootMountPoint")

    if not root_mount_point:
        libcalamares.utils.warning("No mount point for root partition")
        return (_("No mount point for root partition"),
                _("globalstorage does not contain a \"rootMountPoint\" key."))
    if not os.path.exists(root_mount_point):
        libcalamares.utils.warning("Bad root mount point \"{}\"".format(root_mount_point))
        return (_("Bad mount point for root partition"),
                _("rootMountPoint is \"{}\", which does not exist.".format(root_mount_point)))

    if libcalamares.job.configuration.get("unpack", None) is None:
        libcalamares.utils.warning("No *unpack* key in job configuration.")
        return (_("Bad unpackfs configuration"),
                _("There is no configuration information."))

    supported_filesystems = get_supported_filesystems()

    # Bail out before we start when there are obvious problems
    #   - unsupported filesystems
    #   - non-existent sources
    #   - missing tools for specific FS
    for entry in libcalamares.job.configuration["unpack"]:
        source = os.path.abspath(entry["source"])
        sourcefs = entry["sourcefs"]

        if sourcefs not in supported_filesystems:
            libcalamares.utils.warning("The filesystem for \"{}\" ({}) is not supported by your current kernel".format(source, sourcefs))
            libcalamares.utils.warning(" ... modprobe {} may solve the problem".format(sourcefs))
            return (_("Bad unpackfs configuration"),
                    _("The filesystem for \"{}\" ({}) is not supported by your current kernel").format(source, sourcefs))
        if not os.path.exists(source):
            libcalamares.utils.warning("The source filesystem \"{}\" does not exist".format(source))
            return (_("Bad unpackfs configuration"),
                    _("The source filesystem \"{}\" does not exist").format(source))
        if sourcefs == "squashfs":
            if shutil.which("unsquashfs") is None:
                libcalamares.utils.warning("Failed to find unsquashfs")

                return (_("Bad unpackfs configuration"),
                        _("Failed to find unsquashfs, make sure you have the squashfs-tools package installed.") +
                        " " + _("Failed to unpack image \"{}\"").format(source))

    unpack = list()

    is_first = True
    for entry in libcalamares.job.configuration["unpack"]:
        source = os.path.abspath(entry["source"])
        sourcefs = entry["sourcefs"]
        destination = os.path.abspath(root_mount_point + entry["destination"])

        if not os.path.isdir(destination) and sourcefs != "file":
            libcalamares.utils.warning(("The destination \"{}\" in the target system is not a directory").format(destination))
            if is_first:
                return (_("Bad unpackfs configuration"),
                        _("The destination \"{}\" in the target system is not a directory").format(destination))
            else:
                libcalamares.utils.debug(".. assuming that the previous targets will create that directory.")

        unpack.append(UnpackEntry(source, sourcefs, destination))
        # Optional settings
        if entry.get("exclude", None):
            unpack[-1].exclude = entry["exclude"]
        if entry.get("excludeFile", None):
            unpack[-1].excludeFile = entry["excludeFile"]
        unpack[-1].weight = extract_weight(entry)

        is_first = False

    repair_root_permissions(root_mount_point)
    try:
        unpackop = UnpackOperation(unpack)
        return unpackop.run()
    finally:
        repair_root_permissions(root_mount_point)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
#
# === This file is part of Calamares - <https://calamares.io> ===
#
#   SPDX-FileCopyrightText: 2014 Aurélien Gâteau <agateau@kde.org>
#   SPDX-FileCopyrightText: 2014 Anke Boersma <demm@kaosx.us>
#   SPDX-FileCopyrightText: 2014 Daniel Hillenbrand <codeworkx@bbqlinux.org>
#   SPDX-FileCopyrightText: 2014 Benjamin Vaudour <benjamin.vaudour@yahoo.fr>
#   SPDX-FileCopyrightText: 2014-2019 Kevin Kofler <kevin.kofler@chello.at>
#   SPDX-FileCopyrightText: 2015-2018 Philip Mueller <philm@manjaro.org>
#   SPDX-FileCopyrightText: 2016-2017 Teo Mrnjavac <teo@kde.org>
#   SPDX-FileCopyrightText: 2017 Alf Gaida <agaida@siduction.org>
#   SPDX-FileCopyrightText: 2017-2019 Adriaan de Groot <groot@kde.org>
#   SPDX-FileCopyrightText: 2017 Gabriel Craciunescu <crazy@frugalware.org>
#   SPDX-FileCopyrightText: 2017 Ben Green <Bezzy1999@hotmail.com>
#   SPDX-FileCopyrightText: 2021 Neal Gompa <ngompa13@gmail.com>
#   SPDX-License-Identifier: GPL-3.0-or-later
#
#   Calamares is Free Software: see the License-Identifier above.
#

import fileinput
import os
import re
import shutil
import subprocess

import libcalamares

from libcalamares.utils import check_target_env_call

import gettext

_ = gettext.translation("calamares-python",
                        localedir=libcalamares.utils.gettext_path(),
                        languages=libcalamares.utils.gettext_languages(),
                        fallback=True).gettext

# This is the sanitizer used all over to tidy up filenames
# to make identifiers (or to clean up names to make filenames).
file_name_sanitizer = str.maketrans(" /()", "_-__")


def pretty_name():
    return _("Install bootloader.")


def get_uuid():
    """
    Checks and passes 'uuid' to other routine.

    :return:
    """
    partitions = libcalamares.globalstorage.value("partitions")

    for partition in partitions:
        if partition["mountPoint"] == "/":
            libcalamares.utils.debug("Root partition uuid: \"{!s}\"".format(partition["uuid"]))
            return partition["uuid"]

    return ""


def get_kernel_line(kernel_type):
    """
    Passes 'kernel_line' to other routine based on configuration file.

    :param kernel_type:
    :return:
    """
    if kernel_type == "fallback":
        if "fallbackKernelLine" in libcalamares.job.configuration:
            return libcalamares.job.configuration["fallbackKernelLine"]
        else:
            return " (fallback)"
    else:
        if "kernelLine" in libcalamares.job.configuration:
            return libcalamares.job.configuration["kernelLine"]
        else:
            return ""


def get_zfs_root():
    """
    Looks in global storage to find the zfs root

    :return: A string containing the path to the zfs root or None if it is not found
    """

    zfs = libcalamares.globalstorage.value("zfsDatasets")

    if not zfs:
        libcalamares.utils.warning("Failed to locate zfs dataset list")
        return None

    # Find the root dataset
    for dataset in zfs:
        try:
            if dataset["mountpoint"] == "/":
                return dataset["zpool"] + "/" + dataset["dsName"]
        except KeyError:
            # This should be impossible
            libcalamares.utils.warning("Internal error handling zfs dataset")
            raise

    return None


def is_btrfs_root(partition):
    """ Returns True if the partition object refers to a btrfs root filesystem

    :param partition: A partition map from global storage
    :return: True if btrfs and root, False otherwise
    """
    return partition["mountPoint"] == "/" and partition["fs"] == "btrfs"


def is_zfs_root(partition):
    """ Returns True if the partition object refers to a zfs root filesystem

    :param partition: A partition map from global storage
    :return: True if zfs and root, False otherwise
    """
    return partition["mountPoint"] == "/" and partition["fs"] == "zfs"


def have_program_in_target(program : str):
    """Returns @c True if @p program is in path in the target"""
    return libcalamares.utils.target_env_call(["/usr/bin/which", program]) == 0


def get_kernel_params(uuid):
    # Configured kernel parameters (default "quiet"), if plymouth installed, add splash
    # screen parameter and then "rw".
    kernel_params = libcalamares.job.configuration.get("kernelParams", ["quiet"])
    if have_program_in_target("plymouth"):
        kernel_params.append("splash")
    kernel_params.append("rw")

    use_systemd_naming = have_program_in_target("dracut") or (libcalamares.utils.target_env_call(["/usr/bin/grep", "-q", "^HOOKS.*systemd", "/etc/mkinitcpio.conf"]) == 0)

    partitions = libcalamares.globalstorage.value("partitions")

    swap_uuid = ""
    swap_outer_mappername = None
    swap_outer_uuid = None


    cryptdevice_params = []

    # Take over swap settings:
    #  - unencrypted swap partition sets swap_uuid
    #  - encrypted root sets cryptdevice_params
    for partition in partitions:
        if partition["fs"] == "linuxswap" and not partition.get("claimed", None):
            # Skip foreign swap
            continue
        has_luks = "luksMapperName" in partition
        if partition["fs"] == "linuxswap" and not has_luks:
            swap_uuid = partition["uuid"]

        if partition["fs"] == "linuxswap" and has_luks:
            swap_outer_mappername = partition["luksMapperName"]
            swap_outer_uuid = partition["luksUuid"]

        if partition["mountPoint"] == "/" and has_luks:
            if use_systemd_naming:
                cryptdevice_params = [f"rd.luks.uuid={partition['luksUuid']}"]
            else:
                cryptdevice_params = [f"cryptdevice=UUID={partition['luksUuid']}:{partition['luksMapperName']}"]
            cryptdevice_params.append(f"root=/dev/mapper/{partition['luksMapperName']}")

    # btrfs and zfs handling
    for partition in partitions:
        # If a btrfs root subvolume wasn't set, it means the root is directly on the partition
        # and this option isn't needed
        if is_btrfs_root(partition):
            btrfs_root_subvolume = libcalamares.globalstorage.value("btrfsRootSubvolume")
            if btrfs_root_subvolume:
                kernel_params.append("rootflags=subvol=" + btrfs_root_subvolume)

        # zfs needs to be told the location of the root dataset
        if is_zfs_root(partition):
            zfs_root_path = get_zfs_root()
            if zfs_root_path is not None:
                kernel_params.append("root=ZFS=" + zfs_root_path)
            else:
                # Something is really broken if we get to this point
                libcalamares.utils.warning("Internal error handling zfs dataset")
                raise Exception("Internal zfs data missing, please contact your distribution")

    if cryptdevice_params:
        kernel_params.extend(cryptdevice_params)
    else:
        kernel_params.append("root=UUID={!s}".format(uuid))

    if swap_uuid:
        kernel_params.append("resume=UUID={!s}".format(swap_uuid))

    if use_systemd_naming and swap_outer_uuid:
        kernel_params.append(f"rd.luks.uuid={swap_outer_uuid}")

    if swap_outer_mappername:
        kernel_params.append(f"resume=/dev/mapper/{swap_outer_mappername}")

    return kernel_params


def create_systemd_boot_conf(installation_root_path, efi_dir, uuid, kernel, kernel_version):
    """
    Creates systemd-boot configuration files based on given parameters.

    :param installation_root_path: A string containing the absolute path to the root of the installation
    :param efi_dir: A string containing the path to the efi dir relative to the root of the installation
    :param uuid: A string containing the UUID of the root volume
    :param kernel: A string containing the path to the kernel relative to the root of the installation
    :param kernel_version: The kernel version string
    """

    libcalamares.utils.debug(f"Configuring kernel version {kernel_version}")

    # get the machine-id
    with open(os.path.join(installation_root_path, "etc", "machine-id"), 'r') as machineid_file:
        machine_id = machineid_file.read().rstrip('\n')

    # Ensure the directory exists
    machine_dir = os.path.join(installation_root_path + efi_dir, machine_id)
    os.makedirs(machine_dir, exist_ok=True)

    # Call kernel-install for each kernel
    libcalamares.utils.target_env_process_output(["kernel-install",
                                                  "add",
                                                  kernel_version,
                                                  os.path.join("/", kernel)])


def create_loader(loader_path, installation_root_path):
    """
    Writes configuration for loader.

    :param loader_path: The absolute path to the loader.conf file
    :param installation_root_path: The path to the root of the target installation
    """

    # get the machine-id
    with open(os.path.join(installation_root_path, "etc", "machine-id"), 'r') as machineid_file:
        machine_id = machineid_file.read().rstrip('\n')

    try:
        loader_entries = libcalamares.job.configuration["loaderEntries"]
    except KeyError:
        libcalamares.utils.debug("No aditional loader entries found in config")
        loader_entries = []
        pass

    lines = [f"default {machine_id}*"]

    lines.extend(loader_entries)

    with open(loader_path, 'w') as loader_file:
        for line in lines:
            loader_file.write(line + "\n")


class SuffixIterator(object):
    """
    Wrapper for one of the "generator" classes below to behave like
    a proper Python iterator. The iterator is initialized with a
    maximum number of attempts to generate a new suffix.
    """

    def __init__(self, attempts, generator):
        self.generator = generator
        self.attempts = attempts
        self.counter = 0

    def __iter__(self):
        return self

    def __next__(self):
        self.counter += 1
        if self.counter <= self.attempts:
            return self.generator.next()
        raise StopIteration


class serialEfi(object):
    """
    EFI Id generator that appends a serial number to the given name.
    """

    def __init__(self, name):
        self.name = name
        # So the first call to next() will bump it to 0
        self.counter = -1

    def next(self):
        self.counter += 1
        if self.counter > 0:
            return "{!s}{!s}".format(self.name, self.counter)
        else:
            return self.name


def render_in_base(value, base_values, length=-1):
    """
    Renders @p value in base-N, where N is the number of
    items in @p base_values. When rendering, use the items
    of @p base_values (e.g. use "0123456789" to get regular decimal
    rendering, or "ABCDEFGHIJ" for letters-as-numbers 'encoding').

    If length is positive, pads out to at least that long with
    leading "zeroes", whatever base_values[0] is.
    """
    if value < 0:
        raise ValueError("Cannot render negative values")
    if len(base_values) < 2:
        raise ValueError("Insufficient items for base-N rendering")
    if length < 1:
        length = 1
    digits = []
    base = len(base_values)
    while value > 0:
        place = value % base
        value = value // base
        digits.append(base_values[place])
    while len(digits) < length:
        digits.append(base_values[0])
    return "".join(reversed(digits))


class randomEfi(object):
    """
    EFI Id generator that appends a random 4-digit hex number to the given name.
    """

    def __init__(self, name):
        self.name = name
        # So the first call to next() will bump it to 0
        self.counter = -1

    def next(self):
        self.counter += 1
        if self.counter > 0:
            import random
            v = random.randint(0, 65535)  # 16 bits
            return "{!s}{!s}".format(self.name, render_in_base(v, "0123456789ABCDEF", 4))
        else:
            return self.name


class phraseEfi(object):
    """
    EFI Id generator that appends a random phrase to the given name.
    """
    words = ("Sun", "Moon", "Mars", "Soyuz", "Falcon", "Kuaizhou", "Gaganyaan")

    def __init__(self, name):
        self.name = name
        # So the first call to next() will bump it to 0
        self.counter = -1

    def next(self):
        self.counter += 1
        if self.counter > 0:
            import random
            desired_length = 1 + self.counter // 5
            v = random.randint(0, len(self.words) ** desired_length)
            return "{!s}{!s}".format(self.name, render_in_base(v, self.words))
        else:
            return self.name


def get_efi_suffix_generator(name):
    """
    Handle EFI bootloader Ids with ${<something>} for suffix-processing.
    """
    if "${" not in name:
        raise ValueError("Misplaced call to get_efi_suffix_generator, no ${}")
    if not name.endswith("}"):
        raise ValueError("Misplaced call to get_efi_suffix_generator, no trailing ${}")
    if name.count("${") > 1:
        raise ValueError("EFI ID {!r} contains multiple generators".format(name))
    import re
    prefix, generator_name = re.match("(.*)\${([^}]*)}$", name).groups()
    if generator_name not in ("SERIAL", "RANDOM", "PHRASE"):
        raise ValueError("EFI suffix {!r} is unknown".format(generator_name))

    generator = None
    if generator_name == "SERIAL":
        generator = serialEfi(prefix)
    elif generator_name == "RANDOM":
        generator = randomEfi(prefix)
    elif generator_name == "PHRASE":
        generator = phraseEfi(prefix)
    if generator is None:
        raise ValueError("EFI suffix {!r} is unsupported".format(generator_name))

    return generator


def change_efi_suffix(efi_directory, bootloader_id):
    """
    Returns a label based on @p bootloader_id that is usable within
    @p efi_directory. If there is a ${<something>} suffix marker
    in the given id, tries to generate a unique label.
    """
    if bootloader_id.endswith("}") and "${" in bootloader_id:
        # Do 10 attempts with any suffix generator
        g = SuffixIterator(10, get_efi_suffix_generator(bootloader_id))
    else:
        # Just one attempt
        g = [bootloader_id]

    for candidate_name in g:
        if not os.path.exists(os.path.join(efi_directory, candidate_name)):
            return candidate_name
    return bootloader_id


def efi_label(efi_directory):
    """
    Returns a sanitized label, possibly unique, that can be
    used within @p efi_directory.
    """
    if "efiBootloaderId" in libcalamares.job.configuration:
        efi_bootloader_id = change_efi_suffix(efi_directory, libcalamares.job.configuration["efiBootloaderId"])
    else:
        branding = libcalamares.globalstorage.value("branding")
        efi_bootloader_id = branding["bootloaderEntryName"]

    return efi_bootloader_id.translate(file_name_sanitizer)


def efi_word_size():
    # get bitness of the underlying UEFI
    try:
        sysfile = open("/sys/firmware/efi/fw_platform_size", "r")
        efi_bitness = sysfile.read(2)
    except Exception:
        # if the kernel is older than 4.0, the UEFI bitness likely isn't
        # exposed to the userspace so we assume a 64 bit UEFI here
        efi_bitness = "64"
    return efi_bitness


def efi_boot_next():
    """
    Tell EFI to definitely boot into the just-installed
    system next time.
    """
    boot_mgr = libcalamares.job.configuration["efiBootMgr"]
    boot_entry = None
    efi_bootvars = subprocess.check_output([boot_mgr], universal_newlines=True)
    for line in efi_bootvars.split('\n'):
        if not line:
            continue
        words = line.split()
        if len(words) >= 2 and words[0] == "BootOrder:":
            boot_entry = words[1].split(',')[0]
            break
    if boot_entry:
        subprocess.call([boot_mgr, "-n", boot_entry])


def get_kernels(installation_root_path):
    """
    Gets a list of kernels and associated values for each kernel.  This will work as is for many distros.
    If not, it should be safe to modify it to better support your distro

    :param installation_root_path: A string with the absolute path to the root of the installation

    Returns a list of 3-tuples

    Each 3-tuple contains the kernel, kernel_type and kernel_version
    """
    try:
        kernel_search_path = libcalamares.job.configuration["kernelSearchPath"]
    except KeyError:
        libcalamares.utils.warning("No kernel pattern found in configuration, using '/usr/lib/modules'")
        kernel_search_path = "/usr/lib/modules"
        pass

    kernel_list = []

    try:
        kernel_pattern = libcalamares.job.configuration["kernelPattern"]
    except KeyError:
        libcalamares.utils.warning("No kernel pattern found in configuration, using 'vmlinuz'")
        kernel_pattern = "vmlinuz"
        pass

    # find all the installed kernels
    for root, dirs, files in os.walk(os.path.join(installation_root_path, kernel_search_path.lstrip('/'))):
        for file in files:
            if re.search(kernel_pattern, file):
                rel_root = os.path.relpath(root, installation_root_path)
                kernel_list.append((os.path.join(rel_root, file), "default", os.path.basename(root)))

    return kernel_list


def install_clr_boot_manager():
    """
    Installs clr-boot-manager as the bootloader for EFI systems
    """
    libcalamares.utils.debug("Bootloader: clr-boot-manager")

    installation_root_path = libcalamares.globalstorage.value("rootMountPoint")
    kernel_config_path = os.path.join(installation_root_path, "etc", "kernel")
    os.makedirs(kernel_config_path, exist_ok=True)
    cmdline_path = os.path.join(kernel_config_path, "cmdline")

    # Get the kernel params
    uuid = get_uuid()
    kernel_params = " ".join(get_kernel_params(uuid))

    # Write out the cmdline file for clr-boot-manager
    with open(cmdline_path, "w") as cmdline_file:
        cmdline_file.write(kernel_params)

    check_target_env_call(["clr-boot-manager", "update"])


def install_systemd_boot(efi_directory):
    """
    Installs systemd-boot as bootloader for EFI setups.

    :param efi_directory:
    """
    libcalamares.utils.debug("Bootloader: systemd-boot")
    installation_root_path = libcalamares.globalstorage.value("rootMountPoint")
    install_efi_directory = installation_root_path + efi_directory
    uuid = get_uuid()
    loader_path = os.path.join(install_efi_directory,
                               "loader",
                               "loader.conf")
    subprocess.call(["bootctl",
                     "--path={!s}".format(install_efi_directory),
                     "install"])

    # Get the kernel params and write them to /etc/kernel/cmdline
    # This file is used by kernel-install
    kernel_params = " ".join(get_kernel_params(uuid))
    kernel_cmdline_path = os.path.join(installation_root_path, "etc", "kernel")
    os.makedirs(kernel_cmdline_path, exist_ok=True)
    with open(os.path.join(kernel_cmdline_path, "cmdline"), "w") as cmdline_file:
        cmdline_file.write(kernel_params)

    for (kernel, kernel_type, kernel_version) in get_kernels(installation_root_path):
        create_systemd_boot_conf(installation_root_path,
                                 efi_directory,
                                 uuid,
                                 kernel,
                                 kernel_version)

    create_loader(loader_path, installation_root_path)


def get_grub_efi_parameters():
    """
    Returns a 3-tuple of suitable parameters for GRUB EFI installation,
    depending on the host machine architecture. The return is
        - target name
        - grub.efi name
        - boot.efi name
    all three are strings. May return None if there is no suitable
    set for the current machine. May return unsuitable values if the
    host architecture is unknown (e.g. defaults to x86_64).
    """
    import platform
    efi_bitness = efi_word_size()
    cpu_type = platform.machine()

    if efi_bitness == "32":
        # Assume all 32-bitters are legacy x86
        return "i386-efi", "grubia32.efi", "bootia32.efi"
    elif efi_bitness == "64" and cpu_type == "aarch64":
        return "arm64-efi", "grubaa64.efi", "bootaa64.efi"
    elif efi_bitness == "64" and cpu_type == "loongarch64":
        return "loongarch64-efi", "grubloongarch64.efi", "bootloongarch64.efi"
    elif efi_bitness == "64":
        # If it's not ARM, must by AMD64
        return "x86_64-efi", "grubx64.efi", "bootx64.efi"
    libcalamares.utils.warning(
        "Could not find GRUB parameters for bits {b} and cpu {c}".format(b=repr(efi_bitness), c=repr(cpu_type)))
    return None


def run_grub_mkconfig(partitions, output_file):
    """
    Runs grub-mkconfig in the target environment

    :param partitions: The partitions list from global storage
    :param output_file: A string containing the path to the generating grub config file
    :return:
    """

    # zfs needs an environment variable set for grub-mkconfig
    if any([is_zfs_root(partition) for partition in partitions]):
        check_target_env_call(["sh", "-c", "ZPOOL_VDEV_NAME_PATH=1 " +
                               libcalamares.job.configuration["grubMkconfig"] + " -o " + output_file])
    else:
        # The input file /etc/default/grub should already be filled out by the
        # grubcfg job module.
        check_target_env_call([libcalamares.job.configuration["grubMkconfig"], "-o", output_file])


def run_grub_install(fw_type, partitions, efi_directory):
    """
    Runs grub-install in the target environment

    :param fw_type: A string which is "efi" for UEFI installs.  Any other value results in a BIOS install
    :param partitions: The partitions list from global storage
    :param efi_directory: The path of the efi directory relative to the root of the install
    :return:
    """

    is_zfs = any([is_zfs_root(partition) for partition in partitions])

    # zfs needs an environment variable set for grub
    if is_zfs:
        check_target_env_call(["sh", "-c", "echo ZPOOL_VDEV_NAME_PATH=1 >> /etc/environment"])

    if fw_type == "efi":
        assert efi_directory is not None
        efi_bootloader_id = efi_label(efi_directory)
        efi_target, efi_grub_file, efi_boot_file = get_grub_efi_parameters()

        if is_zfs:
            check_target_env_call(["sh", "-c", "ZPOOL_VDEV_NAME_PATH=1 " + libcalamares.job.configuration["grubInstall"]
                                   + " --target=" + efi_target + " --efi-directory=" + efi_directory
                                   + " --bootloader-id=" + efi_bootloader_id + " --force"])
        else:
            check_target_env_call([libcalamares.job.configuration["grubInstall"],
                                   "--target=" + efi_target,
                                   "--efi-directory=" + efi_directory,
                                   "--bootloader-id=" + efi_bootloader_id,
                                   "--force"])
    else:
        assert efi_directory is None
        if libcalamares.globalstorage.value("bootLoader") is None:
            return

        boot_loader = libcalamares.globalstorage.value("bootLoader")
        if boot_loader["installPath"] is None:
            return

        if is_zfs:
            check_target_env_call(["sh", "-c", "ZPOOL_VDEV_NAME_PATH=1 "
                                   + libcalamares.job.configuration["grubInstall"]
                                   + " --target=i386-pc --recheck --force "
                                   + boot_loader["installPath"]])
        else:
            check_target_env_call([libcalamares.job.configuration["grubInstall"],
                                   "--target=i386-pc",
                                   "--recheck",
                                   "--force",
                                   boot_loader["installPath"]])


def install_grub(efi_directory, fw_type):
    """
    Installs grub as bootloader, either in pc or efi mode.

    :param efi_directory:
    :param fw_type:
    """
    # get the partition from global storage
    partitions = libcalamares.globalstorage.value("partitions")
    if not partitions:
        libcalamares.utils.warning(_("Failed to install grub, no partitions defined in global storage"))
        return

    if fw_type == "efi":
        libcalamares.utils.debug("Bootloader: grub (efi)")
        installation_root_path = libcalamares.globalstorage.value("rootMountPoint")
        install_efi_directory = installation_root_path + efi_directory

        if not os.path.isdir(install_efi_directory):
            os.makedirs(install_efi_directory)

        efi_bootloader_id = efi_label(efi_directory)

        efi_target, efi_grub_file, efi_boot_file = get_grub_efi_parameters()

        run_grub_install(fw_type, partitions, efi_directory)

        # VFAT is weird, see issue CAL-385
        install_efi_directory_firmware = (vfat_correct_case(
            install_efi_directory,
            "EFI"))
        if not os.path.exists(install_efi_directory_firmware):
            os.makedirs(install_efi_directory_firmware)

        # there might be several values for the boot directory
        # most usual they are boot, Boot, BOOT

        install_efi_boot_directory = (vfat_correct_case(
            install_efi_directory_firmware,
            "boot"))
        if not os.path.exists(install_efi_boot_directory):
            os.makedirs(install_efi_boot_directory)

        # Workaround for some UEFI firmwares
        fallback = "installEFIFallback"
        libcalamares.utils.debug("UEFI Fallback: " + str(libcalamares.job.configuration.get(fallback, "<unset>")))
        if libcalamares.job.configuration.get(fallback, True):
            libcalamares.utils.debug("  .. installing '{!s}' fallback firmware".format(efi_boot_file))
            efi_file_source = os.path.join(install_efi_directory_firmware,
                                           efi_bootloader_id,
                                           efi_grub_file)
            efi_file_target = os.path.join(install_efi_boot_directory, efi_boot_file)

            shutil.copy2(efi_file_source, efi_file_target)
    else:
        libcalamares.utils.debug("Bootloader: grub (bios)")
        run_grub_install(fw_type, partitions, None)

    run_grub_mkconfig(partitions, libcalamares.job.configuration["grubCfg"])


def install_secureboot(efi_directory):
    """
    Installs the secureboot shim in the system by calling efibootmgr.
    """
    efi_bootloader_id = efi_label(efi_directory)

    installation_root_path = libcalamares.globalstorage.value("rootMountPoint")
    install_efi_directory = installation_root_path + efi_directory

    if efi_word_size() == "64":
        install_efi_bin = "shimx64.efi"
    elif efi_word_size() == "32":
        install_efi_bin = "shimia32.efi"
    else:
        libcalamares.utils.warning(f"Unknown efi word size of {efi_word_size()} found")
        return None

    # Copied, roughly, from openSUSE's install script,
    # and pythonified. *disk* is something like /dev/sda,
    # while *drive* may return "(disk/dev/sda,gpt1)" ..
    # we're interested in the numbers in the second part
    # of that tuple.
    efi_drive = subprocess.check_output([
        libcalamares.job.configuration["grubProbe"],
        "-t", "drive", "--device-map=", install_efi_directory]).decode("ascii")
    efi_disk = subprocess.check_output([
        libcalamares.job.configuration["grubProbe"],
        "-t", "disk", "--device-map=", install_efi_directory]).decode("ascii")

    efi_drive_partition = efi_drive.replace("(", "").replace(")", "").split(",")[1]
    # Get the first run of digits from the partition
    efi_partition_number = None
    c = 0
    start = None
    while c < len(efi_drive_partition):
        if efi_drive_partition[c].isdigit() and start is None:
            start = c
        if not efi_drive_partition[c].isdigit() and start is not None:
            efi_partition_number = efi_drive_partition[start:c]
            break
        c += 1
    if efi_partition_number is None:
        raise ValueError("No partition number found for %s" % install_efi_directory)

    subprocess.call([
        libcalamares.job.configuration["efiBootMgr"],
        "-c",
        "-w",
        "-L", efi_bootloader_id,
        "-d", efi_disk,
        "-p", efi_partition_number,
        "-l", install_efi_directory + "/" + install_efi_bin])

    efi_boot_next()

    # The input file /etc/default/grub should already be filled out by the
    # grubcfg job module.
    check_target_env_call([libcalamares.job.configuration["grubMkconfig"],
                           "-o", os.path.join(efi_directory, "EFI",
                                              efi_bootloader_id, "grub.cfg")])


def vfat_correct_case(parent, name):
    for candidate in os.listdir(parent):
        if name.lower() == candidate.lower():
            return os.path.join(parent, candidate)
    return os.path.join(parent, name)


def efi_partitions(efi_boot_path):
    """
    The (one) partition mounted on @p efi_boot_path, or an empty list.
    """
    return [p for p in libcalamares.globalstorage.value("partitions") if p["mountPoint"] == efi_boot_path]


def update_refind_config(efi_directory, installation_root_path):
    """
    :param efi_directory: The path to the efi directory relative to the root
    :param installation_root_path: The path to the root of the installation
    """
    try:
        kernel_list = libcalamares.job.configuration["refindKernelList"]
    except KeyError:
        libcalamares.utils.warning('refindKernelList not set.  Skipping updating refind.conf')
        return

    # Update the config in the file
    for line in fileinput.input(installation_root_path + efi_directory + "/EFI/refind/refind.conf", inplace=True):
        line = line.strip()
        if line.startswith("#extra_kernel_version_strings") or line.startswith("extra_kernel_version_strings"):
            line = line.lstrip("#")
            for kernel in kernel_list:
                if kernel not in line:
                    line += "," + kernel
        print(line)


def install_refind(efi_directory):
    try:
        installation_root_path = libcalamares.globalstorage.value("rootMountPoint")
    except KeyError:
        libcalamares.utils.warning('Global storage value "rootMountPoint" missing')

    install_efi_directory = installation_root_path + efi_directory
    uuid = get_uuid()
    kernel_params = " ".join(get_kernel_params(uuid))
    conf_path = os.path.join(installation_root_path, "boot/refind_linux.conf")

    check_target_env_call(["refind-install"])

    with open(conf_path, "r") as refind_file:
        filedata = [x.strip() for x in refind_file.readlines()]

    with open(conf_path, 'w') as refind_file:
        for line in filedata:
            if line.startswith('"Boot with standard options"'):
                line = f'"Boot with standard options"    "{kernel_params}"'
            elif line.startswith('"Boot to single-user mode"'):
                line = f'"Boot to single-user mode"    "{kernel_params}" single'
            refind_file.write(line + "\n")

    update_refind_config(efi_directory, installation_root_path)


def prepare_bootloader(fw_type):
    """
    Prepares bootloader.
    Based on value 'efi_boot_loader', it either calls systemd-boot
    or grub to be installed.

    :param fw_type:
    :return:
    """

    # Get the boot loader selection from global storage if it is set in the config file
    try:
        gs_name = libcalamares.job.configuration["efiBootLoaderVar"]
        if libcalamares.globalstorage.contains(gs_name):
            efi_boot_loader = libcalamares.globalstorage.value(gs_name)
        else:
            libcalamares.utils.warning(
                f"Specified global storage value not found in global storage")
            return None
    except KeyError:
        # If the conf value for using global storage is not set, use the setting from the config file.
        try:
            efi_boot_loader = libcalamares.job.configuration["efiBootLoader"]
        except KeyError:
            if fw_type == "efi":
                libcalamares.utils.warning("Configuration missing both efiBootLoader and efiBootLoaderVar on an EFI "
                                           "system, bootloader not installed")
                return
            else:
                pass

    # If the user has selected not to install bootloader, bail out here
    if efi_boot_loader.casefold() == "none":
        libcalamares.utils.debug("Skipping bootloader installation since no bootloader was selected")
        return None

    efi_directory = libcalamares.globalstorage.value("efiSystemPartition")

    if efi_boot_loader == "clr-boot-manager":
        if fw_type != "efi":
            # Grub has to be installed first on non-EFI systems
            install_grub(efi_directory, fw_type)
        install_clr_boot_manager()
    elif efi_boot_loader == "systemd-boot" and fw_type == "efi":
        install_systemd_boot(efi_directory)
    elif efi_boot_loader == "sb-shim" and fw_type == "efi":
        install_secureboot(efi_directory)
    elif efi_boot_loader == "refind" and fw_type == "efi":
        install_refind(efi_directory)
    elif efi_boot_loader == "grub" or fw_type != "efi":
        install_grub(efi_directory, fw_type)
    else:
        libcalamares.utils.debug("WARNING: the combination of "
                                 "boot-loader '{!s}' and firmware '{!s}' "
                                 "is not supported.".format(efi_boot_loader, fw_type))


def run():
    """
    Starts procedure and passes 'fw_type' to other routine.

    :return:
    """

    fw_type = libcalamares.globalstorage.value("firmwareType")

    if libcalamares.globalstorage.value("bootLoader") is None and fw_type != "efi":
        libcalamares.utils.warning("Non-EFI system, and no bootloader is set.")
        return None

    partitions = libcalamares.globalstorage.value("partitions")
    if fw_type == "efi":
        efi_system_partition = libcalamares.globalstorage.value("efiSystemPartition")
        esp_found = [p for p in partitions if p["mountPoint"] == efi_system_partition]
        if not esp_found:
            libcalamares.utils.warning("EFI system, but nothing mounted on {!s}".format(efi_system_partition))
            return None

    try:
        prepare_bootloader(fw_type)
    except subprocess.CalledProcessError as e:
        libcalamares.utils.warning(str(e))
        libcalamares.utils.debug("stdout:" + str(e.stdout))
        libcalamares.utils.debug("stderr:" + str(e.stderr))
        return (_("Bootloader installation error"),
                _("The bootloader could not be installed. The installation command <pre>{!s}</pre> returned error "
                  "code {!s}.")
                .format(e.cmd, e.returncode))

    return None



#!/usr/bin/env python3

import os
import re
import subprocess
import shutil
from subprocess import CalledProcessError

import libcalamares
from libcalamares.utils import gettext_path, gettext_languages

import gettext
from pathlib import Path

_translation = gettext.translation("calamares-python",
                                   localedir=gettext_path(),
                                   languages=gettext_languages(),
                                   fallback=True)
_ = _translation.gettext
_n = _translation.ngettext

name = "Setup pacman and configure mirrors"
user_output = False


def pretty_name():
    return _(name)


def pretty_status_message():
    return 'Updating mirrorlists and initializing keyrings'


def line_cb(line):
    """
    Writes every line to the debug log and displays it in calamares
    :param line: The line of output text from the command
    """
    if user_output:
        global custom_status_message
        custom_status_message = line.strip()
    libcalamares.utils.debug(line)


def is_package_file_modified(package_name, search_string, filename):
    if Path(filename).exists():
        diff_re = re.compile(f".*{search_string}.*checksum.*")

        pacman_output = list()
        try:
            libcalamares.utils.host_env_process_output(['pacman', '-Qkk', package_name], pacman_output)
            for line in pacman_output:
                if diff_re.match(line):
                    return True
        except CalledProcessError:
            libcalamares.utils.warning("Failed to process existing Arch mirrorlist, continuing...")

    return False


def generate_mirrorlist(command, run_command):
    if Path(command).exists():
        try:
            libcalamares.utils.host_env_process_output(run_command, line_cb)
            return True
        except CalledProcessError:
            libcalamares.utils.warning(f"{command} missing")

    return False


def update_mirrorlist_online(mountpoint):
    relative_command = "/usr/bin/rate-mirrors"
    eos_relative_command = "/usr/bin/eos-rankmirrors"
    fallback_command = ["/usr/bin/create-ml"]
    arch_mirrorlist_filename = "/etc/pacman.d/mirrorlist"
    eos_mirrorlist_filename = "/etc/pacman.d/endeavouros-mirrorlist"

    # Check if the Arch mirrorlist on the host has been modified
    use_existing_mirrorlist = is_package_file_modified('pacman-mirrorlist', 'mirrorlist', arch_mirrorlist_filename)

    # Generate the Arch mirrorlist if needed
    if use_existing_mirrorlist:
        libcalamares.utils.debug('Using existing Arch mirrorlist from host')
    else:
        if not generate_mirrorlist(relative_command, [relative_command, "--allow-root", "--save=/etc/pacman.d/mirrorlist",
                                                      "arch", "--max-delay=3600"]):
            # Use the fallback mirrorlist
            libcalamares.utils.warning("Generating fallback mirrorlist")
            libcalamares.utils.host_env_process_output(fallback_command, line_cb)

    # Now copy the mirrorlist to the target
    install_file(mountpoint, arch_mirrorlist_filename)

    # Check if the EOS mirrorlist on the host has been modified
    use_existing_mirrorlist = is_package_file_modified('endeavouros-mirrorlist', 'endeavouros', eos_mirrorlist_filename)

    # Generate the EOS mirrorlist if needed
    if use_existing_mirrorlist:
        libcalamares.utils.debug('Using existing EOS mirrorlist from host')
    else:
        if not generate_mirrorlist(eos_relative_command, [eos_relative_command]):
            libcalamares.utils.warning(f"Failed to generate EOS mirrorlist")

    # Now copy the mirrorlist to the target
    install_file(mountpoint, eos_mirrorlist_filename)



def update_mirrorlist_offline(mountpoint):
    command = ["/usr/bin/create-ml", "--offline", "--save"]
    arch_mirrorlist_filename = "/etc/pacman.d/mirrorlist"
    eos_mirrorlist_filename = "/etc/pacman.d/endeavouros-mirrorlist"

    use_existing_mirrorlist = is_package_file_modified('pacman-mirrorlist', 'mirrorlist', arch_mirrorlist_filename)

    if use_existing_mirrorlist:
        libcalamares.utils.debug('Using existing Arch mirrorlist from host')
    else:
        libcalamares.utils.host_env_process_output(command, line_cb)

    install_file(mountpoint, arch_mirrorlist_filename)
    install_file(mountpoint, eos_mirrorlist_filename)


def install_file(mountpoint, filename):
    try:
        target = mountpoint + filename
        libcalamares.utils.debug(f'Installing {filename} to {target}')
        Path(target).parent.mkdir(parents=True, exist_ok=True)
        shutil.copy2(filename, target)
    except:
        libcalamares.utils.warning(f"Failed to install {filename} to target")


def run():
    """
    Runs a script passing in the appropriate options

    """
    global user_output
    user_output = libcalamares.job.configuration.get("userOutput", False)

    try:
        root_mountpoint = libcalamares.globalstorage.value("rootMountPoint")
    except KeyError:
        return f"Missing root mountpoint, aborting"

    # If hasInternet is missing from global storage, it means the user selected an offline install
    if libcalamares.globalstorage.contains("hasInternet"):
        update_mirrorlist_online(root_mountpoint)
        try:
            libcalamares.utils.host_env_process_output(["pacman", "-Sy", "--noconfirm", "archlinux-keyring", "endeavouros-keyring"],
                                                       line_cb)
        except CalledProcessError:
            libcalamares.utils.warning(f"Failed to update keyring on host")
    else:
        update_mirrorlist_offline(root_mountpoint)
        try:
            libcalamares.utils.host_env_process_output(["pacman-key", "--init"], line_cb)
            libcalamares.utils.host_env_process_output(["pacman-key", "--populate", "archlinux", "endeavouros"], line_cb)
        except CalledProcessError:
            libcalamares.utils.warning(f"Failed to update keyring on host")

    return None



#!/usr/bin/env python3

import os
import subprocess
import shutil
import time

import libcalamares
from libcalamares.utils import gettext_path, gettext_languages

import gettext

_translation = gettext.translation("calamares-python",
                                   localedir=gettext_path(),
                                   languages=gettext_languages(),
                                   fallback=True)
_ = _translation.gettext
_n = _translation.ngettext

custom_status_message = None
status_update_time = 0


class PacmanError(Exception):
    """Exception raised when the call to pacman returns a non-zero exit code

    Attributes:
        message -- explanation of the error
    """

    def __init__(self, message):
        self.message = message


def pretty_name():
    return _("Install base system")


def pretty_status_message():
    if custom_status_message is not None:
        return custom_status_message


def line_cb(line):
    """
    Writes every line to the debug log and displays it in calamares
    :param line: The line of output text from the command
    """
    global custom_status_message
    custom_status_message = line.strip()
    libcalamares.utils.debug("pacstrap: " + line.strip())
    libcalamares.job.setprogress(0)


def run_in_host(command, line_func):
    proc = subprocess.Popen(command, stdout=subprocess.PIPE, stderr=subprocess.STDOUT, universal_newlines=True,
                            bufsize=1)
    for line in proc.stdout:
        if line.strip():
            line_func(line)
    proc.wait()
    if proc.returncode != 0:
        raise PacmanError("Failed to run pacman")


def run():
    """
    Installs the base system packages and copies files post-installation

    """
    root_mount_point = libcalamares.globalstorage.value("rootMountPoint")

    if not root_mount_point:
        return ("No mount point for root partition in globalstorage",
                "globalstorage does not contain a \"rootMountPoint\" key, "
                "doing nothing")

    if not os.path.exists(root_mount_point):
        return ("Bad mount point for root partition in globalstorage",
                "globalstorage[\"rootMountPoint\"] is \"{}\", which does not "
                "exist, doing nothing".format(root_mount_point))

    if libcalamares.job.configuration:
        if "basePackages" in libcalamares.job.configuration:
            base_packages = libcalamares.job.configuration["basePackages"]
        else:
            return "Package List Missing", "Cannot continue without list of packages to install"
    else:
        return "No configuration found", "Aborting due to missing configuration"

    # run the pacstrap
    pacstrap_command = ["/etc/calamares/scripts/pacstrap_calamares", "-c", root_mount_point] + base_packages

    try:
        run_in_host(pacstrap_command, line_cb)
    except subprocess.CalledProcessError as cpe:
        return "Failed to run pacstrap", "Pacstrap failed with error {!s}".format(cpe.stderr)
    except PacmanError as pe:
        return "Failed to run pacstrap", format(pe)

    # Initialize the keyring
    try:
        libcalamares.utils.target_env_process_output(["pacman-key", "--init"], None)
        libcalamares.utils.target_env_process_output(["pacman-key", "--populate", "archlinux", "endeavouros"], None)
    except CalledProcessError:
        libcalamares.utils.warning(f"Failed to update keyring on target")

    # copy files post install
    if "postInstallFiles" in libcalamares.job.configuration:
        files_to_copy = libcalamares.job.configuration["postInstallFiles"]
        for source_file in files_to_copy:
            if os.path.exists(source_file):
                try:
                    libcalamares.utils.debug("Copying file {!s}".format(source_file))
                    dest = os.path.normpath(root_mount_point + source_file)
                    os.makedirs(os.path.dirname(dest), exist_ok=True)
                    shutil.copy2(source_file, dest)
                except Exception as e:
                    libcalamares.utils.warning("Failed to copy file {!s}, error {!s}".format(source_file, e))

    libcalamares.globalstorage.insert("online", True)

    libcalamares.job.setprogress(1.0)

    return None



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
#
# === This file is part of Calamares - <https://calamares.io> ===
#
#   SPDX-FileCopyrightText: 2014 Rohan Garg <rohan@kde.org>
#   SPDX-FileCopyrightText: 2015 Philip Müller <philm@manjaro.org>
#   SPDX-FileCopyrightText: 2016 David McKinney <mckinney@subgraph.com>
#   SPDX-FileCopyrightText: 2016 Kevin Kofler <kevin.kofler@chello.at>
#   SPDX-FileCopyrightText: 2017 Alf Gaida <agaida@siduction.org>
#   SPDX-FileCopyrightText: 2017 2019, Adriaan de Groot <groot@kde.org>
#   SPDX-License-Identifier: GPL-3.0-or-later
#
#   Calamares is Free Software: see the License-Identifier above.
#

import libcalamares

import inspect
import os
import shutil


import gettext
_ = gettext.translation("calamares-python",
                        localedir=libcalamares.utils.gettext_path(),
                        languages=libcalamares.utils.gettext_languages(),
                        fallback=True).gettext


def pretty_name():
    return _("Configuring initramfs.")


def copy_initramfs_hooks(partitions, root_mount_point):
    """
    Copies initramfs hooks so they are picked up by update-initramfs

    :param partitions:
    :param root_mount_point:
    """
    encrypt_hook = False
    unencrypted_separate_boot = False

    for partition in partitions:
        if partition["mountPoint"] == "/" and "luksMapperName" in partition:
            encrypt_hook = True

        if (partition["mountPoint"] == "/boot"
                and "luksMapperName" not in partition):
            unencrypted_separate_boot = True

    if encrypt_hook:
        target = "{!s}/usr/share/initramfs-tools/hooks/encrypt_hook".format(
                    root_mount_point)

        # Find where this module is installed
        _filename = inspect.getframeinfo(inspect.currentframe()).filename
        _path = os.path.dirname(os.path.abspath(_filename))

        if unencrypted_separate_boot:
            shutil.copy2(
                os.path.join(_path, "encrypt_hook_nokey"),
                target
                )
        else:
            shutil.copy2(
                os.path.join(_path, "encrypt_hook"),
                target
                )
        os.chmod(target, 0o755)


def run():
    """
    Calls routine with given parameters to configure initramfs

    :return:
    """
    partitions = libcalamares.globalstorage.value("partitions")
    root_mount_point = libcalamares.globalstorage.value("rootMountPoint")

    if not partitions:
        libcalamares.utils.warning("partitions is empty, {!s}".format(partitions))
        return (_("Configuration Error"),
                _("No partitions are defined for <pre>{!s}</pre> to use." ).format("initramfscfg"))
    if not root_mount_point:
        libcalamares.utils.warning("rootMountPoint is empty, {!s}".format(root_mount_point))
        return (_("Configuration Error"),
                _("No root mount point is given for <pre>{!s}</pre> to use." ).format("initramfscfg"))

    copy_initramfs_hooks(partitions, root_mount_point)

    return None



#!/usr/bin/env python3

import subprocess

import libcalamares
from libcalamares.utils import gettext_path, gettext_languages
from subprocess import CalledProcessError
import gettext

_translation = gettext.translation("calamares-python",
                                   localedir=gettext_path(),
                                   languages=gettext_languages(),
                                   fallback=True)
_ = _translation.gettext
_n = _translation.ngettext

custom_status_message = "Detecting hardware and installing drivers"
name = "Hardware detection"

def pretty_name():
    return _(name)


def pretty_status_message():
    if custom_status_message is not None:
        return custom_status_message


def line_cb(line):
    """
    Writes every line to the debug log and displays it in calamares
    :param line: The line of output text from the command
    """
    global custom_status_message
    custom_status_message = line.strip()
    libcalamares.utils.debug(line)


def get_cpu_info():
    cpu_model = "unknown"
    cpu_vendor = "unknown"
    try:
        with open("/proc/cpuinfo", "r") as cpu_file:
            for line in cpu_file:
                if line.strip().startswith("vendor_id"):
                    cpu_vendor = line.split(":")[1].strip()
                if line.strip().startswith("model name"):
                    cpu_model = line.split(":")[1].strip()
    except KeyError:
        libcalamares.utils.warning("Failed to get CPU information")

    return cpu_vendor, cpu_model;


def install_ucode(vendor):
    package = None
    if vendor == 'GenuineIntel':
        package = 'intel-ucode'
    elif vendor == 'AuthenticAMD':
        package = 'amd-ucode'

    if package:
        try:
            libcalamares.utils.target_env_process_output(["pacman", "-Sy", "--noconfirm", package], line_cb)
        except CalledProcessError:
            libcalamares.utils.warning(f"Failed to install {package}")
    else:
        libcalamares.utils.warning(f'Vendor {vendor} has no know ucode package.  Skipping ucode install...')


def remove_ucode(vendor):
    packages = list()
    if vendor != 'GenuineIntel':
        packages.append('intel-ucode')
    elif vendor != 'AuthenticAMD':
        packages.append('amd-ucode')

    if packages:
        try:
            libcalamares.utils.target_env_process_output(["pacman", "-Rcn", "--noconfirm"] + packages, callback=line_cb)
        except CalledProcessError:
            libcalamares.utils.warning(f"Failed to remove {packages}")


def run_command(command):
    try:
        libcalamares.utils.target_env_process_output(command, line_cb)
    except CalledProcessError:
        libcalamares.utils.warning(f"Failed to run {command}")


def run():
    vendor, model = get_cpu_info()
    hw_tool = '/usr/bin/eos-hwtool'

    if libcalamares.globalstorage.contains("hasInternet"):
        install_ucode(vendor)
        run_command([hw_tool, '--iso', '--no32', '--install-recommended'])
    else:
        remove_ucode(vendor)
        run_command([hw_tool, '--iso', '--install-recommended', '--packagedir=/usr/share/packages', '--nvidia-only'])
        run_command([hw_tool, '--iso', '--purge'])
        run_command([hw_tool, '--enable-services'])

    return None



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
#
# === This file is part of Calamares - <https://calamares.io> ===
#
#   SPDX-FileCopyrightText: 2014 Philip Müller <philm@manjaro.org>
#   SPDX-FileCopyrightText: 2014 Teo Mrnjavac <teo@kde.org>
#   SPDX-FileCopyrightText: 2017 Alf Gaida <agaida@siduction.org>
#   SPDX-FileCopyrightText: 2018-2019 Adriaan de Groot <groot@kde.org>
#   SPDX-FileCopyrightText: 2022 shivanandvp <shivanandvp@rebornos.org>
#   SPDX-License-Identifier: GPL-3.0-or-later
#
#   Calamares is Free Software: see the License-Identifier above.

import libcalamares


import gettext
_ = gettext.translation("calamares-python",
                        localedir=libcalamares.utils.gettext_path(),
                        languages=libcalamares.utils.gettext_languages(),
                        fallback=True).gettext


def pretty_name():
    return _("Configure systemd units")


def systemctl(units):
    """
    For each entry in @p units, run "systemctl <action> <name>",
    where each unit is a mapping of unit name, action, and a flag.

    Returns a failure message, or None if this was successful.
    Units that are not mandatory have their failures suppressed
    silently.
    """

    for unit in units:
        if isinstance(unit, str):
            name = unit
            action = "enable"
            mandatory = False
        else:
            if "name" not in unit:
                libcalamares.utils.error("The key 'name' is missing from the mapping {_unit!s}. Continuing to the next unit.".format(_unit=str(unit)))
                continue 
            name = unit["name"]
            action = unit.get("action", "enable")
            mandatory = unit.get("mandatory", False)

        exit_code = libcalamares.utils.target_env_call(
            ['systemctl', action, name]
        )

        if exit_code != 0:
            libcalamares.utils.warning(
                "Cannot {} systemd unit {}".format(action, name)
            )
            libcalamares.utils.warning(
                "systemctl {} call in chroot returned error code {}".format(action, exit_code)
            )
            if mandatory:
                title = _("Cannot modify unit")
                diagnostic = _("<code>systemctl {_action!s}</code> call in chroot returned error code {_exit_code!s}.").format(_action=action, _exit_code=exit_code)
                description = _("Cannot {_action!s} systemd unit <code>{_name!s}</code>.").format(_action=action, _name=name)
                return (
                    title,
                    description + " " + diagnostic
                )
    return None


def run():
    """
    Setup systemd units
    """
    cfg = libcalamares.job.configuration

    return_value = systemctl(
        cfg.get("units", [])
    )
    if return_value is not None:
        return return_value

    return None



#!/usr/bin/env python3

import os
import subprocess
import shutil

import libcalamares
from libcalamares.utils import gettext_path, gettext_languages

import gettext

_translation = gettext.translation("calamares-python",
                                   localedir=gettext_path(),
                                   languages=gettext_languages(),
                                   fallback=True)
_ = _translation.gettext
_n = _translation.ngettext

custom_status_message = None
name = "Run script"
user_output = False


def pretty_name():
    return _(name)


def pretty_status_message():
    if custom_status_message is not None:
        return custom_status_message


def line_cb(line):
    """
    Writes every line to the debug log and displays it in calamares
    :param line: The line of output text from the command
    """
    if user_output:
        global custom_status_message
        custom_status_message = line.strip()
    libcalamares.utils.debug(line)


def run():
    """
    Runs a script passing in the appropriate options

    """

    if not libcalamares.job.configuration:
        return "No configuration found", "Aborting due to missing configuration"

    if "name" in libcalamares.job.configuration:
        global name
        name = libcalamares.job.configuration["name"]

    if "scriptPath" in libcalamares.job.configuration:
        script_path = libcalamares.job.configuration["scriptPath"]
    else:
        return "Configuration Error", "No script defined in config"

    online = libcalamares.job.configuration.get("isOnline", False)
    run_in_target = libcalamares.job.configuration.get("runInTarget", False)
    include_root = libcalamares.job.configuration.get("includeRoot", False)
    include_user = libcalamares.job.configuration.get("includeUser", False)
    include_raw_user = libcalamares.job.configuration.get("includeRawUser", False)
    global user_output
    user_output = libcalamares.job.configuration.get("userOutput", False)

    try:
        gs_name = libcalamares.job.configuration["gsName"]
    except KeyError:
        gs_name = None

    # build the parameter list
    command = [script_path]
    if online and libcalamares.globalstorage.contains("hasInternet"):
        if libcalamares.globalstorage.value("hasInternet"):
            command.append("--online")

    if gs_name is not None:
        try:
            command.append("--bootloader=" + libcalamares.globalstorage.value(gs_name))
        except KeyError:
            return f"Key missing", f"Failed to find {gs_name} in global storage"

    if include_root:
        root_mount_point = libcalamares.globalstorage.value("rootMountPoint")

        if not root_mount_point:
            return ("No mount point for root partition in globalstorage",
                    "globalstorage does not contain a \"rootMountPoint\" key, "
                    "doing nothing")

        if not os.path.exists(root_mount_point):
            return ("Bad mount point for root partition in globalstorage",
                    "globalstorage[\"rootMountPoint\"] is \"{}\", which does not "
                    "exist, doing nothing".format(root_mount_point))

        command.append("--root=" + root_mount_point)

    if include_user or include_raw_user:
        user = libcalamares.globalstorage.value("username")
        if user:
            if include_user:
                command.append(f"--user={user}")
            else:
                command.append(user)
        else:
            return "Username missing", "Username not available in global storage"

    try:
        if run_in_target:
            libcalamares.utils.target_env_process_output(command, line_cb)
        else:
            libcalamares.utils.host_env_process_output(command, line_cb)
    except subprocess.CalledProcessError as cpe:
        return "Failed to run script", "The script failed with error {!s}".format(cpe.stderr)

    return None
